from __future__ import annotations

import typing
from .. import transforms

if typing.TYPE_CHECKING:
    import polars as pl
    import tooltree
    import tooltime


def plot_treemap(
    metrics: pl.DataFrame,
    metric: typing.Literal['volume', 'open_interest'],
    *,
    timescale: str | None = None,
    start_timestamp: tooltime.Timestamp | None = None,
    end_timestamp: tooltime.Timestamp | None = None,
    levels: list[str] = ['category', 'series_title', 'contract_ticker'],
    max_children: int = 100,
    min_child_fraction: float = 0.0001,
    max_depth: int = 3,
    plot_kwargs: dict[str, typing.Any] | None = None,
) -> tooltree.TreemapPlot:
    import tooltree
    import polars as pl

    if metric == 'volume':
        if (
            timescale is None
            and start_timestamp is None
            and end_timestamp is None
        ):
            timescale = '1d'
        if start_timestamp is None and end_timestamp is None:
            end_timestamp = metrics['timestamp'].max()  # type: ignore
        data = transforms.get_metrics_subrange(
            metrics,
            start_time=start_timestamp,
            end_time=end_timestamp,
            duration=timescale,
        )
    elif metric == 'open_interest':
        if start_timestamp is None:
            start_timestamp = metrics['timestamp'].max()  # type: ignore
        data = metrics.filter(timestamp=start_timestamp)
    else:
        raise Exception('invalid metric: ' + metric)

    if plot_kwargs is None:
        plot_kwargs = {}

    return tooltree.plot_treemap(
        data,
        levels=levels,
        metric=metric,
        root=_get_title(data=data, metric=metric, timescale=timescale),
        metric_format=_get_dollar_format(),
        height=700,
        max_depth=max_depth,
        max_children=max_children,
        min_child_fraction=min_child_fraction,
        **plot_kwargs,
    )


def _get_dollar_format() -> dict[str, typing.Any]:
    return {'order_of_magnitude': True, 'decimals': 1, 'prefix': '$'}


def _get_title(data: pl.DataFrame, metric: str, timescale: str | None) -> str:
    if metric == 'open_interest':
        timerange_str = '[' + str(data['timestamp'].max())[:10] + ']'
        return 'Kalshi ' + metric.title() + ' ' + timerange_str

    if timescale is None:
        raise Exception('timescale must be specified for volume metric')
    timerange_str = (
        '['
        + str(data['min_timestamp'][0])[:10]
        + ', '
        + str(data['max_timestamp'][0])[:10]
        + ']'
    )
    if timescale == '1d':
        timerange_str = '[' + str(data['timestamp'].max())[:10] + ']'
        title = 'Kalshi ' + metric.title() + ' Over Past 24H'
    elif timescale == 'all':
        title = 'Kalshi Lifetime ' + metric.title()
    else:
        title = 'Kalshi ' + metric.title() + ' Over Past ' + timescale
    return title + ' ' + timerange_str



from __future__ import annotations

import polars as pl
from . import transforms
from . import categories


def load_kalshi_data(
    *,
    timescale: str = '1d',
    trim_finalized: bool = True,
    add_subcategories: bool = True,
) -> pl.DataFrame:
    import absorb

    raw_metrics = absorb.query('kalshi.metrics')
    metadata = absorb.query('kalshi.metadata')
    metrics = _join_series_metadata(raw_metrics, metadata).select(
        timestamp='timestamp',
        series_ticker=pl.col.series_ticker.fill_null('Unknown'),
        contract_ticker='ticker_name',
        category=pl.col.category.fill_null('Unknown'),
        series_title=pl.col.series_title.fill_null('Unknown'),
        open_interest='open_interest',
        volume='daily_volume',
        status='status',
    )

    if trim_finalized:
        metrics = transforms.trim_finalized_markets(metrics)

    metrics = _refactor_raw_categories(metrics)
    if add_subcategories:
        metrics = categories.populate_subcategories(metrics)
        metrics = metrics.with_columns(
            pl.col.subcategory.fill_null(pl.col.series_title)
        )

    if timescale == '1d':
        return metrics
    else:
        return transforms.downsample_metrics(metrics, timescale=timescale)


def _join_series_metadata(
    kalshi_metrics: pl.DataFrame, kalshi_metadata: pl.DataFrame
) -> pl.DataFrame:
    """add series columns

    columns
    - series_ticker
    - category
    - series_title
    """
    joined_series = (
        kalshi_metrics.select(pl.col.report_ticker.unique())
        .join(
            kalshi_metadata.unique('series_ticker').select(
                'series_ticker', 'series_title', 'category'
            ),
            left_on='report_ticker',
            right_on='series_ticker',
            how='left',
            coalesce=False,
        )
        .join(
            kalshi_metadata.unique('series_ticker').select(
                'series_ticker', 'series_title', 'category'
            ),
            left_on='KX' + pl.col.report_ticker.str.to_uppercase(),
            right_on='series_ticker',
            how='left',
            coalesce=False,
        )
        .select(
            'report_ticker',
            series_ticker=pl.coalesce('series_ticker', 'series_ticker_right'),
            series_title=pl.coalesce('series_title', 'series_title_right'),
            category=pl.coalesce('category', 'category_right'),
        )
        .join(
            _get_manual_series_metadata(),
            left_on='report_ticker',
            right_on='series_ticker',
            how='left',
            coalesce=False,
        )
        .select(
            'report_ticker',
            series_ticker=pl.coalesce('series_ticker', 'series_ticker_right'),
            series_title=pl.coalesce('series_title_right', 'series_title'),
            category=pl.coalesce('category_right', 'category'),
        )
    )

    return kalshi_metrics.join(joined_series, on='report_ticker', how='left')


def _get_manual_series_metadata() -> pl.DataFrame:
    # (series_ticker, category, series_title)
    manual_series_metadata = [
        #
        # financials
        ('NASDAQ100D', 'NASDAQ100D', 'Financials'),
        ('NASDAQ100W', 'NASDAQ100W', 'Financials'),
        ('NASDAQ100DU', 'NASDAQ100DU', 'Financials'),
        ('NASDAQ100M', 'NASDAQ100M', 'Financials'),
        ('INXD', 'S&P 500 daily', 'Financials'),
        ('INXW', 'S&P 500 weekly', 'Financials'),
        ('INXDU', 'S&P 500 daily up/down', 'Financials'),
        ('INXM', 'S&P 500 monthly', 'Financials'),
        #
        # sports
        ('KXNFLSPREAD', 'NFL Football Game, Point Spread', 'Sports'),
        ('KXNFLTOTAL', 'NFL Football Game, Point Total', 'Sports'),
        ('KXNFLANYTD', 'NFL Football Game', 'Sports'),
        ('KXNCAAFSPREAD', 'NCAA Football Game', 'Sports'),
        ('KXNCAAFTOTAL', 'NCAA Football Game, Point Total', 'Sports'),
        ('KXNFLWINS', 'NFL Football Game, Who wins?', 'Sports'),
        ('KXNFLFIRSTTD', 'NFL Football Game, First Touchdown', 'Sports'),
        ('KXNFL2TD', 'NFL Football Game, 2 or more touchdowns', 'Sports'),
        #
        # politics
        ('KXPENNWIN', 'Pennsylvania win', 'Politics'),
        ('KXMAYORSF', 'San Francisco Mayor', 'Politics'),
        ('KXECADVANTAGE', 'Electoral College advantage', 'Politics'),
        ('KXLASTSTATECALL24', 'Last state called in 2024', 'Politics'),
        ('KXELECTION24RCP', 'Election 24 RCP', 'Politics'),
        ('KXELECTION24538', 'Election 24 538', 'Politics'),
        ('KXDJTPAWIN', 'Trump winning PA and the Presidency', 'Politics'),
        (
            'KXKHBLUEWALL',
            'Harris wins but loses any Blue Wall state',
            'Politics',
        ),
        (
            'KXDJTBLUEWALL',
            'Trump wins but loses any Blue Wall state',
            'Politics',
        ),
        ('KXKHMI', 'Harris wins but loses Michigan', 'Politics'),
        ('KXELECTIONPACALL', 'PA call', 'Politics'),
        ('KXELECTIONMICALL', 'MI call', 'Politics'),
        ('KXELECTIONGACALL', 'GA call', 'Politics'),
        ('KXEXITPOLLLMEN', 'Latino men 2024', 'Politics'),
        ('KXEXITPOLLMEN', 'Men in 2024', 'Politics'),
        ('KXEXITPOLLWOMEN', 'Women in 2024', 'Politics'),
        ('KXKHUNMWOMEN', 'Unmarried women 2024', 'Politics'),
        ('KXKHCGRADS', 'College graduates in 2024', 'Politics'),
        ('KXDJTMEN', 'Men in 2024', 'Politics'),
        ('KXDJTASIAN', 'Asians in 2024', 'Politics'),
        ('KXPAWINPOPVOTE', 'PA win popular vote nationally', 'Politics'),
        ('KXECPOPVOTE', 'EC/Pop Vote 2024', 'Politics'),
        ('KX3RDPARTY', '3rd Party 2% of vote', 'Politics'),
        ('KXRFKSHARE', 'RFK vote share', 'Politics'),
        ('KXDJTLGBT', 'LGBT voters 2024', 'Politics'),
        ('KXEXITPOLLLATE', 'Late deciders', 'Politics'),
        ('KXEXITPOLLWWOMEN', 'White women 2024', 'Politics'),
        ('KXDJTLATINO', 'Latinos 2024', 'Politics'),
        ('KXDJTNOGRAD', 'Non-college graduates 2024', 'Politics'),
        ('KXKHWOMEN', 'Women in 2024', 'Politics'),
        ('KX538CALL', '538 call 2024', 'Politics'),
        ('KXSILVERCALL', 'Silver call 2024', 'Politics'),
        ('KXOHDISTRICTS', 'Ohio redistricting referendum', 'Politics'),
    ]
    return pl.DataFrame(
        manual_series_metadata,
        schema=['series_ticker', 'series_title', 'category'],
        orient='row',
    )


def _refactor_raw_categories(metrics: pl.DataFrame) -> pl.DataFrame:
    rename_category = {
        'Entertainment': 'Culture',
        'Science and Technology': 'STEM',
    }

    merge_category = {
        'Companies': 'Financials',
        'Elections': 'Politics',
        'Mentions': 'Politics',
    }

    merge_as_subcategory = {
        'Social': 'Culture',
        'Health': 'STEM',
        'Climate and Weather': 'STEM',
        'Transportation': 'STEM',
        'Education': 'STEM',
    }

    return (
        metrics.with_columns(pl.col.category.replace(rename_category))
        .with_columns(pl.col.category.replace(merge_category))
        .with_columns(
            subcategory=pl.when(
                pl.col.category.is_in(merge_as_subcategory.keys())
            )
            .then('category')
            .otherwise(None)
        )
        .with_columns(pl.col.category.replace(merge_as_subcategory))
    )


def _get_uncategorized_kalshi_volume(df: pl.DataFrame) -> pl.DataFrame:
    return (
        df.filter(series_ticker='Unknown')
        .group_by('report_ticker')
        .agg(pl.sum('daily_volume'))
        .sort('daily_volume', descending=True)
    )



from __future__ import annotations


def get_category_colors() -> dict[str, str]:
    return {
        'World': '#0077BB',
        'STEM': '#AA4499',
        'Sports': '#228833',
        'Economics': '#EE7733',
        'Politics': '#CC3311',
        'Financials': '#009988',
        'Crypto': '#33BBEE',
        'Culture': '#EE3377',
        'Unknown': '#BBBBBB',
    }



from __future__ import annotations

import typing
from . import colors as colors_module

if typing.TYPE_CHECKING:
    import polars as pl
    import plotly.graph_objects as go  # type: ignore


def plot_metric_by_category(
    data: pl.DataFrame, metric_column: str, metric_name: str
) -> None:
    import polars as pl
    import plotly.graph_objects as go
    import toolstr

    fig = go.Figure()
    colors = colors_module.get_category_colors()

    project = False

    categories = (
        data.group_by('category')
        .agg(metric=pl.sum(metric_column))
        .sort('metric', descending=True)['category']
    )

    total = data.group_by('timestamp', maintain_order=True).agg(
        pl.sum(metric_column)
    )
    fig.add_trace(
        go.Scatter(
            x=total['timestamp'],
            y=total[metric_column],
            mode='lines',
            name='TOTAL',
            line=dict(color='black', width=5),
            legendgroup='TOTAL',
            customdata=[
                toolstr.format(
                    value, order_of_magnitude=True, decimals=1, prefix='$'
                )
                for value in total[metric_column]
            ],
            line_simplify=False,
            hovertemplate='TOTAL' + ': %{customdata}<extra></extra>',
        )
    )

    for category in categories:
        category_data = data.filter(category=category)
        x = category_data['timestamp']
        y = category_data[metric_column]
        color = colors.get(category, '#e377c2')

        if not project:
            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=y,
                    mode='lines',
                    name=category,
                    line=dict(color=color, width=3),
                    legendgroup=category,
                    customdata=[
                        toolstr.format(
                            value,
                            order_of_magnitude=True,
                            decimals=1,
                            prefix='$',
                        )
                        for value in y
                    ],
                    line_simplify=False,
                    hovertemplate=category + ': %{customdata}<extra></extra>',
                )
            )
        else:
            fig.add_trace(
                go.Scatter(
                    x=x[:-1],
                    y=y[:-1],
                    mode='lines',
                    name=category,
                    line=dict(color=color, width=3),
                    legendgroup=category,
                    customdata=[
                        toolstr.format(
                            value,
                            order_of_magnitude=True,
                            decimals=1,
                            prefix='$',
                        )
                        for value in y[:-1]
                    ],
                    line_simplify=False,
                    hovertemplate=category + ': %{customdata}<extra></extra>',
                )
            )
            fig.add_trace(
                go.Scatter(
                    x=x[-2:],
                    y=y[-2:],
                    mode='lines',
                    line=dict(color=color, width=3, dash='5 3'),
                    marker=dict(size=6),
                    hoverinfo='skip',
                    showlegend=False,
                    legendgroup=category,
                )
            )
            fig.add_trace(
                go.Scatter(
                    x=x[-1:],
                    y=y[-1:],
                    mode='markers',
                    line=dict(color=color, width=3, dash='dot'),
                    marker=dict(size=6),
                    name=category,
                    showlegend=False,
                    legendgroup=category,
                )
            )

    grid_style = dict(
        showgrid=True,  # Show/hide gridlines
        gridwidth=1,  # Width of gridlines
        gridcolor='rgba(128, 128, 128, 0.2)',  # Color with transparency
        griddash='5 3',
    )
    label_style = {'size': 18, 'color': '#000000'}

    fig.update_layout(
        margin=dict(t=55, b=0, l=0, r=0, pad=0),
        title={
            'text': 'Kalshi ' + metric_name + ' By Category',
            'x': 0.5,  # Centers the title
            'xanchor': 'center',
            'font': {'size': 24, 'color': '#000000'},
        },
        yaxis=dict(
            title={'text': metric_name, 'font': label_style, 'standoff': 24},
            range=[
                -total[metric_column][-7:].max() * 0.07,  # type: ignore
                total[metric_column][-7:].max() * 1.1,  # type: ignore
            ],
            fixedrange=False,
            tickfont=label_style,
            tickprefix='$',
            **grid_style,
        ),
        height=600,
        hovermode='x unified',
        xaxis=dict(
            range=['2025-03-01', data['timestamp'].max()],
            title={'font': label_style},
            tickfont=label_style,
            tickformatstops=[
                dict(dtickrange=['M12', None], value='%Y'),
                dict(dtickrange=['M1', 'M12'], value='%Y-%m'),
                dict(dtickrange=[None, 'M1'], value='%Y-%m-%d'),
            ],
            rangeslider=dict(visible=True, thickness=0.1),
            type='date',
            **grid_style,
        ),
        legend=dict(
            orientation='v',
            yanchor='top',
            y=1,
            xanchor='left',
            x=1.02,
            font=label_style,
        ),
        template='plotly_white',
    )
    fig.update_xaxes(hoverformat='%Y-%m-%d')
    hires_config = {
        'toImageButtonOptions': {
            'format': 'png',
            'filename': 'kalshi-volume',
            'height': 600,
            'width': 1000,
            'scale': 2,
        },
        'displayModeBar': False,
    }

    fig.show(config=hires_config)



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import polars as pl
    import tooltime


def trim_finalized_markets(metrics: pl.DataFrame) -> pl.DataFrame:
    """Remove all but the first finalized data point for each market"""
    import polars as pl

    return (
        metrics.with_columns(
            first_appearance=pl.struct(
                pl.col.contract_ticker, pl.col.status
            ).is_first_distinct()
        )
        .filter((pl.col.status != 'finalized') | (pl.col.first_appearance))
        .drop('first_appearance')
    )


def get_metrics_subrange(
    metrics: pl.DataFrame,
    *,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    duration: str | None = None,
) -> pl.DataFrame:
    import polars as pl
    import tooltime
    import datetime

    if start_time is None and end_time is None and duration is None:
        raise ValueError('must specify two of start_time, end_time, duration')
    if duration == 'all':
        start_time = metrics['timestamp'].min()  # type: ignore
        end_time = metrics['timestamp'].max()  # type: ignore
    if start_time is None:
        if duration is not None and end_time is not None:
            start_time = tooltime.timestamp_to_seconds(
                end_time
            ) - tooltime.timelength_to_seconds(duration)
        else:
            raise ValueError(
                'must specify two of start_time, end_time, duration'
            )
    if end_time is None:
        if duration is not None and start_time is not None:
            end_time = tooltime.timestamp_to_seconds(
                start_time
            ) + tooltime.timelength_to_seconds(duration)
        else:
            raise ValueError(
                'must specify two of start_time, end_time, duration'
            )

    start_time = tooltime.timestamp_to_datetime(start_time).replace(
        tzinfo=datetime.timezone.utc
    )
    end_time = tooltime.timestamp_to_datetime(end_time).replace(
        tzinfo=datetime.timezone.utc
    )

    filtered = metrics.filter(
        pl.col.timestamp >= start_time, pl.col.timestamp < end_time
    )
    return (
        filtered.group_by('contract_ticker', maintain_order=True)
        .agg(**_get_resample_columns())
        .with_columns(timestamp=pl.lit(start_time))
        .select(
            *metrics.columns,
            end_timestamp=pl.lit(end_time),
            min_timestamp=pl.lit(filtered['timestamp'].min()),
            max_timestamp=pl.lit(filtered['timestamp'].max()),
        )
    )


def get_metrics_by_category(metrics: pl.DataFrame) -> pl.DataFrame:
    import polars as pl

    return metrics.group_by('timestamp', 'category', maintain_order=True).agg(
        pl.col.volume.sum(),
        pl.col.open_interest.sum(),
        pl.col.series_ticker.n_unique().alias('event_types'),
        pl.col.contract_ticker.n_unique().alias('contracts'),
        pl.col.subcategory.n_unique().alias('subcategories'),
    )


def downsample_metrics(
    metrics: pl.DataFrame, timescale: str, *, include_status: bool = False
) -> pl.DataFrame:
    import polars as pl

    if not include_status and 'status' in metrics.columns:
        metrics = metrics.drop('status')

    if timescale == 'all':
        timescale = '100y'

    if timescale.endswith('w'):
        timestamp = (
            pl.col.timestamp.dt.offset_by('1d')
            .dt.truncate(timescale)
            .dt.offset_by('-1d')
        )
    else:
        timestamp = pl.col.timestamp.dt.truncate(timescale)

    resample_columns = {
        k: v
        for k, v in _get_resample_columns().items()
        if k in metrics.columns or k in {'min_time', 'max_time'}
    }
    return (
        metrics.group_by(
            timestamp=timestamp,
            contract_ticker='contract_ticker',
            maintain_order=True,
        )
        .agg(**resample_columns)
        .select(*metrics.columns, 'min_time', 'max_time')
    )


def _get_resample_columns() -> dict[str, pl.Expr]:
    import polars as pl

    return dict(
        series_ticker=pl.col.series_ticker.first(),
        category=pl.col.category.first(),
        subcategory=pl.col.subcategory.first(),
        series_title=pl.col.series_title.first(),
        volume=pl.col.volume.sum(),
        open_interest=pl.col.open_interest.first(),
        status=pl.col.status.unique(),
        min_time=pl.col.timestamp.min(),
        max_time=pl.col.timestamp.max(),
    )


def add_event_slugs(metrics: pl.DataFrame) -> pl.DataFrame:
    all_contracts = (
        metrics.unique(('series_ticker', 'contract_ticker', 'series_title'))
        .select('series_ticker', 'contract_ticker', 'series_title')
        .with_columns(
            n_series_dashes=pl.col.series_ticker.str.count_matches('-'),
            n_contract_dashes=pl.col.contract_ticker.str.count_matches('-'),
        )
    )

    event_slugs = all_contracts.filter(
        pl.col.n_series_dashes + 2 == pl.col.n_contract_dashes,
        pl.col.contract_ticker.str.starts_with(pl.col.series_ticker)
        | pl.col.contract_ticker.str.starts_with('KX' + pl.col.series_ticker),
    ).with_columns(
        event_slug=pl.col.contract_ticker.str.split('-').list.get(1),
    )

    return metrics.join(
        event_slugs.select('contract_ticker', 'event_slug'),
        on='contract_ticker',
        how='left',
    )



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import polars as pl


def print_data_summary(data: pl.DataFrame) -> None:
    import toolstr

    toolstr.print_text_box('Data Summary')
    print(
        '- data range: ['
        + str(data['timestamp'].min())[:10]
        + ', '
        + str(data['timestamp'].max())[:10]
        + ']'
    )
    print('- categories:', data['category'].n_unique())
    print('- subcategories:', toolstr.format(data['subcategory'].n_unique()))
    print('- event types:', toolstr.format(data['series_ticker'].n_unique()))
    print('- contracts:', toolstr.format(data['contract_ticker'].n_unique()))


def get_category_coverage(
    uncategorized: pl.DataFrame, data: pl.DataFrame
) -> pl.DataFrame:
    import datetime
    import polars as pl

    max_timestamp = data['timestamp'].max()
    last_24H = pl.col.timestamp == max_timestamp
    last_7D = pl.col.timestamp >= max_timestamp - datetime.timedelta(days=7)  # type: ignore
    last_30D = pl.col.timestamp >= max_timestamp - datetime.timedelta(days=30)  # type: ignore

    return uncategorized.select(
        (
            1
            - pl.col.volume.filter(last_24H).sum()
            / data.filter(last_24H)['volume'].sum()
        ).alias('last\n24H'),
        (
            1
            - pl.col.volume.filter(last_7D).sum()
            / data.filter(last_7D)['volume'].sum()
        ).alias('last\n7D'),
        (
            1
            - pl.col.volume.filter(last_30D).sum()
            / data.filter(last_30D)['volume'].sum()
        ).alias('last\n30D'),
        (1 - pl.col.volume.sum() / data['volume'].sum()).alias('all\ntime'),
    )


def get_coverage(data: pl.DataFrame) -> pl.DataFrame:
    import polars as pl

    categorized_filter = ~pl.col.category.is_in(['World', 'Unknown'])
    subcategorized_filter = categorized_filter & (
        pl.col.series_title != pl.col.subcategory
    )

    # category coverage
    uncategorized_coverage = get_category_coverage(
        data.filter(~categorized_filter),
        data,
    )

    # subcategory coverage
    unsubcategorized_coverage = get_category_coverage(
        data.filter(~subcategorized_filter),
        data,
    )

    # combine all
    return pl.concat(
        [
            uncategorized_coverage,
            unsubcategorized_coverage,
        ]
    ).insert_column(
        0,
        pl.Series(['categorized', 'subcategorized']),
    )


def print_coverage(metrics: pl.DataFrame) -> None:
    import toolstr

    coverage = get_coverage(metrics)
    toolstr.print_text_box('Volume Label Coverage')
    column_formats = {
        'last\n24H': {'percentage': True, 'decimals': 2},
        'last\n7D': {'percentage': True, 'decimals': 2},
        'last\n30D': {'percentage': True, 'decimals': 2},
        'all\ntime': {'percentage': True, 'decimals': 2},
    }
    toolstr.print_dataframe_as_table(
        coverage, compact=3, column_formats=column_formats
    )



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import polars as pl


subcategory_triggers = {
    'Sports': {
        'Football': [
            '(?i)football',
            'NFL',
            'Super Bowl',
            'Heisman Trophy',
            'NCAAF',
            'AFC',
            'NFC',
            'Big 12',
            'SEC Champion',
            'ACC Champion',
            'Big Ten Champion',
            'SB mention',
            'UFL',
        ],
        'Soccer': [
            '(?i)soccer',
            'MLS',
            'English Premier League Game',
            'UEFA',
            'Club World Club Game',
            '(?i)La Liga',
            '(?i)Premier League',
            'Ligue 1',
            'Serie A',
            '(?i)Bundesliga',
            'Club World Cup',
            "Ballon d'Or",
            'EPL top 4 teams',
            "Men's World Cup",
            'FA cup',
            'Champions League',
            'Europa Leauge',
            'PFA Player of The Year',
            'Club World Club Group',
            'Europa League',
            'Ronaldo',
            'Real Madrid',
        ],
        'Baseball': [
            '(?i)baseball',
            'MLB',
            'World Series',
            'American League East',
            'American League West',
            'American League Central',
            'National League Central',
            'National League East',
            'National League West',
        ],
        'Basketball': [
            '(?i)basketball',
            'NBA',
            r'March Tournament \(M\)',
            r'March Tournament \(W\)',
            'LBJ retire',
            'NIT Champion',
        ],
        'Tennis': [
            'Tennis',
            '(?i)WIMBLEDON',
            'French Open',
            'ATP',
            'WTA',
            'Indian Wells Open',
            'Movistar Chile Open',
            'Rio Open',
            'Laver Cup',
            'Qatar ExxonMobil Open',
            'Djokovic',
        ],
        'Golf': [
            'PGA',
            'US Open',
            'Masters Tournament',
            'The Open Championship',
            'Ryder Cup',
            'Genesis Invitational',
            'LIV Tour',
            'Phoenix Open',
            'Masters Cut',
        ],
        'Martial Arts': ['UFC', 'Boxing'],
        'Hockey': ['Hockey', 'Stanley Cup', 'Conn Smythe'],
        'Automotive': [
            'F1 Race',
            'NASCAR',
            'Indy 500',
            'Formula 1',
            'Verstappen',
        ],
        'Eating': ["(?i)Nathan's Hot Dog"],
        'Chess': ['Chess', 'FIDE'],
        'Cricket': ['Asia Cup Winner', 'IPL Final'],
        'Pickleball': ['pickleball'],
        'Esports': [
            '2025 Mid-Season Invitational',
            'Dota 2',
            'Esports',
            'League of Legends',
        ],
        'Cycling': ['Tour de France'],
        'Running': ['4-Minute Mile'],
        'Frisbee': ['Frisbee'],
    },
    'Culture': {
        'Music': [
            'Spotify',
            'Grammy',
            'Billboard',
            'Eurovision',
            'Album record',
            'pop singer',
            'Top artist',
            'New albums',
            'Artist of the Year',
            'Top songs',
            'Will Taylor Swift release a',
            'Taylor Swift album',
            '(?i)Cardi B',
            'Kendrick Lamar',
            'Justin Bieber',
            'Billie Eilish',
            'CHARLI XCX',
            '(?i)Sabrina Carpenter',
            'Kanye',
            '(?i)Taylor Swift',
            '(?i)Album',
            '(?i)song',
            '(?i)many straight weeks',
            '(?i)doja cat',
            'Top New Artist',
            'Will any new height video receive 100 million views in 7 days',
            'Who will be featured on the life of a show girl',
            'Donda',
            'listen',
            'ACMA',
        ],
        'Trump': ['Trump'],
        'Movies': [
            '(?i)Rotten tomatoes',
            'Oscar',
            'RT',
            'SAG',
            'Netflix movie ranking',
            '(?i)film',
            'Critics Choice Award',
            "Palme d'Or",
            'Ghostbusters',
            'BAFTA',
            'Next Bond actor',
            'Golden Globe for Best Actor in a Drama',
            'Golden Globe for Best Director',
        ],
        'TV': [
            '(?i)Emmy',
            'Netflix TV rank',
            'White Lotus',
            'Jimmy Kimmel release',
            'TV',
            'Anime',
            '(?i)Season',
        ],
        'Video Games': ['Game Awards', 'GTA ', 'Clair Obscur', 'PS6'],
        'Person of the Year': ['Person of the Year'],
        'Podcast': [
            'Rogan',
            'JRE',
            'how many views will taylor swift get on video',
            'Lex Fridman',
            'podcast',
        ],
        'Sports': [
            '(?i)Super Bowl',
            ' BEST TEAM‬$',
            'March Madness',
            'NBA',
            'MLB',
            'head coach',
            'Bronny James',
        ],
        'Apps': ['Free app ranking'],
        'Internet': ['Mr. Beast'],
        'Biden': ['(?i)Biden'],
        'Strike': [
            'Writers strike ends',
            'Actors strike ends',
        ],
        'Relationship': ['Swift Kelce'],
        'Martial Arts': ['UFC', 'Hulk Hogan'],
        'Musical': ['Best Play', 'Best Musical', 'in a Musical'],
    },
    'STEM': {
        'AI': [
            'ChatGPT',
            '(?i)LLM',
            '^Top model$',
            'Unsupervised full self-driving',
            'AGI',
            'Top-ranked AI',
            'IMO AI',
            'OpenAI',
            'AI coding',
            '^Open-source model$',
            'Claude 4 release',
            'Deepseek',
            'AI ',
            'SWE bench',
            'Model hitting thresholds$',
            'GPT',
            '(?i)Grok',
        ],
        'Transportation': ['TSA check-ins this week', 'Waymo'],
        'Space': [
            'Aliens',
            'Starship',
            'NASA',
            'Webb Telescope launch',
            'Exoplanet',
            'Geomagnetic Storm',
            'SpaceX',
            'Major meteor strike',
            'Asteroid',
            'astronaut',
            'Moon',
        ],
        'Health': ['Ozempic', 'Omicron vaccine', 'cancer'],
    },
    'Financials': {
        'Nasdaq': ['(?i)Nasdaq'],
        'S&P': ['S&P', 'SPX', 'SPY', 'SP500'],
        'Tesla': ['(?i)tesla', 'Robotaxi'],
        'Forex': ['USD/JPY', 'EUR/USD', 'EURO/USD', 'Pound/US Dollar'],
        'Earnings mention': [
            '(?i)earnings mention',
            'earnings m$',
            'earnings menti$',
            r'earnings call\?$',
        ],
        'SpaceX': ['SpaceX'],
        'TikTok': ['(?i)TikTok'],
        'Treasury': ['Treasury'],
        'OpenAI': ['(?i)OpenAI'],
        'Coinbase': ['Coinbase'],
        'Nintendo': ['Nintendo'],
        'Commodities': ['WTI oil'],
        'IPO': ['IPOs'],
    },
    'Crypto': {
        'Bitcoin': ['(?i)Bitcoin', 'BTC'],
        'Ethereum': ['(?i)Ethereum', 'ETH'],
        'Solana': ['(?i)Solana', 'SOL'],
        'Doge': ['(?i)Doge'],
        'Shiba': ['(?i)Shiba'],
        'BCH': ['BCH'],
        'XRP': ['XRP'],
        'Avalanche': ['Avalanche'],
        'XLM': ['XLM'],
        'LTC': ['LTC'],
        'Chainlink': ['Chainlink'],
        'Polkadot': ['Polkadot'],
        'Stablecoins': ['USDC', 'USDT', 'UDST', '(?i)stablecoin'],
        'General': [
            'Crypto being positive',
            'New coin launch',
            'Crypto reserve assets',
            'Crypto performance',
            'Next coin in RH',
            'New coin chain',
        ],
    },
    'Economics': {
        'Inflation': ['(?i)inflation', 'CPI'],
        'IPO': ['IPO'],
        'Interest Rates': [
            'Fed ',
            'rate cuts',
            'Large cuts',
            'Rate cut',
            'Zero rates',
        ],
        'Commodities': [
            '(?i)Gas',
            'Oil',
            '(?i)Semiconductor',
            'prices',
            'Price of',
            'Costco hot dog price increase',
            'ISM PMI',
            'Diesel price',
            'SPR use',
            'Used Vehicle Value Index',
        ],
        'Real Estate': [
            '(?i)real estate',
            'home sales',
            'house average sale',
            'Mortgage',
            'Rent',
            'US housing price up monthly',
            'yearly rent growth',
            'NYC rent yearly increase',
        ],
        'Strike': ['strike'],
        'Wealth': [
            'First trillionaire',
            'Musk wealth',
            'Musk trillionaire',
            'Top 3 wealthiest people',
            'Wealthiest person in world',
            'Half-trillionaire',
            'Person wealthiest',
        ],
        'Forex': ['exchange rate'],
        'Trade': [
            '(?i)tariff',
            'Export tariffs',
            'trade deficit',
            'Suez traffic',
            'U.S. BOP',
        ],
        'Employment': [
            'Jobs numbers',
            '(?i)Unemployment',
            'Weekly initial jobless claims',
            'Initial jobless claims',
            'ADP employment change',
            'layoffs',
            'Employment-population ratio',
            'Job revisions',
        ],
        'GDP': ['GDP', '(?i)recession'],
        'Debt': ['debt', 'Trillion dollar coin minted'],
        'Insolvency': ['bankrupt', 'card defaults'],
    },
}


def _get_subcategory_series() -> dict[str, list[str]]:
    return {
        'Basketball': [
            'KXNBAFINALSMVP',
            'KXNEXTTEAMGIANNIS',
            'KXMANTISFREETHROWS',
            'KXNYKCOACH',
            'KXNEXTTEAMLEBRON',
        ],
        'Hockey': ['KXNHLEAST', 'KXNHLWEST', 'KXNHL4NATIONS'],
        'Baseball': [
            'KXMLBNLCPOTY',
        ],
        'Soccer': ['KXUCLROUND'],
        'Football': [
            'KXNEWCOACHNO',
            'KXNEWCOACHDAL',
            'KXNEWCOACHCHIBEARS',
            'KXNEXTTEAMMICAH',
            'KXNEWCOACHNYJ',
            'KXNEWCOACHLV',
            'KXNEWCOACHJAX',
            'KXTUSHPUSH',
            'KXFFQBALLENMAHOMES',
            'KXSTARTCLEBROWNS',
            'KXSTARTIND',
            'KXSTARTNOSAINTS',
            'KXNEWCOACHNE',
            'KXNEXTTEAMMCLAURIN',
        ],
        'Golf': ['KXFIRSTUSOPEN'],
        'Esports': [
            'KXFANATICSGAMESFIRSTPLACE',
            'KXINTERNETINVITATIONAL',
            'KXVCCHAMPIONSPARIS',
        ],
    }


def _get_recategorize_series() -> dict[str, tuple[str, typing.Optional[str]]]:
    return {
        'KXREVSOL': ('Crypto', 'SOL'),
        'KXWTAX': ('Politics', None),
        'KXRAINNOSB': ('STEM', 'Climate and Weather'),
        'KXRAINNY': ('STEM', 'Climate and Weather'),
        'KXHIGHNY0': ('STEM', 'Climate and Weather'),
        'KXNEXTPOPE': ('Culture', 'Religion'),
        'KXNEWPOPEDATE': ('Culture', 'Religion'),
        'KXNEWPOPEROUNDS': ('Culture', 'Religion'),
        'KXNEWPOPECONTINENT': ('Culture', 'Religion'),
        'KXNEWPOPENAME': ('Culture', 'Religion'),
        'KXPOPEVISITNEXT': ('Culture', 'Religion'),
        'KXNEWPOPE': ('Culture', 'Religion'),
        'KXNEWPOPECOUNTRY': ('Culture', 'Religion'),
        'KXRECSS': ('Economics', 'Recession'),
        'KXUSADEALCOUNT': ('Economics', 'Trade'),
        'KXFOOTBALL1001': ('Sports', 'Football'),
        'KXNYCBOROUGHWINBRON': ('Politics', 'NYC Mayor'),
        'KXNYCBOROUGHWINMAN': ('Politics', 'NYC Mayor'),
        'KXHAPPIER': ('Culture', 'Music'),
        'KXFERTILITYSK': ('Economics', 'Demographics'),
        'KXPRCBIRTHS': ('Economics', 'Demographics'),
        'KXCASE7DCN': ('STEM', 'Health'),
        'KXTRUMPVISITSTATES': ('Culture', 'Trump'),
        'KXMONKEYPOX': ('STEM', 'Health'),
        'KXMAERSK': ('Economics', 'Trade'),
        'KXGDPCN': ('Economics', 'GDP'),
        'KXDINENYC': ('STEM', 'Health'),
        'KXARGCENTRALBANK': ('Economics', 'Insolvency'),
        'KXARGDOLLAR': ('Economics', None),
        'KXNYCBOROUGHWINQUEE': ('Politics', 'NYC Mayor'),
        'KXNYCBOROUGHWINBROOK': ('Politics', 'NYC Mayor'),
        'KXMAMDANIBOROUGHS': ('Politics', 'NYC Mayor'),
        'KXSLOAN': ('Economics', 'Debt'),
        'KXSHAKEDURATIONDJTPUTIN': ('Culture', 'Trump'),
        'KXPAHLAVIVISITA': ('Politics', 'Iran'),
        'KXINTERSTELLAR': ('STEM', 'Space'),
        'KXDEBATEMAMDANICUOMO': ('Politics', 'NYC Mayor'),
        'KXSEAICE': ('STEM', 'Climate and Weather'),
        'KXPMLA': ('STEM', 'Climate and Weather'),
        'KXOLYMCANCEL': ('STEM', 'Health'),
        'KXSUEZ': ('Economics', 'Trade'),
        'KXNYCBOROUGHWINSTAT': ('Politics', 'NYC Mayor'),
        'KXINDEXADD': ('Financials', 'S&P 500'),
    }


def populate_subcategories(data: pl.DataFrame) -> pl.DataFrame:
    import polars as pl

    original_data = data
    columns = [
        'series_ticker',
        'contract_ticker',
        'series_title',
        'category',
        'subcategory',
    ]
    data = data.unique(columns).select(columns)

    # trigger keywords
    for category in subcategory_triggers.keys():
        partitions = data.with_columns(
            subcategory_is_null=pl.col.category == category
        ).partition_by('subcategory_is_null', as_dict=True)

        subcategory: str | None
        category_data = partitions[(True,)]
        when_replace: typing.Any = pl.when(False).then(None)
        for subcategory, keywords in subcategory_triggers[category].items():
            filter = (
                pl.col.category == category
            ) & pl.col.series_title.str.contains(keywords[0])
            for keyword in keywords[1:]:
                filter = filter | (
                    (pl.col.category == category)
                    & pl.col.series_title.str.contains(keyword)
                )
            when_replace = when_replace.when(filter).then(pl.lit(subcategory))
        category_data = category_data.with_columns(
            subcategory=when_replace.otherwise('subcategory')
        )
        data = (
            partitions[(False,)]
            .vstack(category_data)
            .drop('subcategory_is_null')
        )

    # direct replacements by subcategory
    when_replace = pl.when(False).then(None)
    subcategory_series = _get_subcategory_series()
    recategorize_series = _get_recategorize_series()
    for series_ticker, (category, subcategory) in recategorize_series.items():
        if subcategory is not None:
            if subcategory not in subcategory_series:
                subcategory_series[subcategory] = []
            subcategory_series[subcategory].append(series_ticker)
    for subcategory, keywords in subcategory_series.items():
        filter = pl.col.series_ticker == keywords[0]
        for keyword in keywords[1:]:
            filter = filter | (pl.col.series_ticker == keyword)
        when_replace = when_replace.when(filter).then(pl.lit(subcategory))
    data = data.with_columns(subcategory=when_replace.otherwise('subcategory'))

    # direct replacements by category
    category_series: dict[str, list[str]] = {}
    for series_ticker, (category, subcategory) in recategorize_series.items():
        category_series.setdefault(category, []).append(series_ticker)
    when_replace = pl
    for subcategory, keywords in category_series.items():
        filter = pl.col.series_ticker == keywords[0]
        for keyword in keywords[1:]:
            filter = filter | (pl.col.series_ticker == keyword)
        when_replace = when_replace.when(filter).then(pl.lit(subcategory))
    data = data.with_columns(category=when_replace.otherwise('category'))

    return original_data.drop('category', 'subcategory').join(
        data.sort('contract_ticker').select(
            'contract_ticker', 'category', 'subcategory'
        ),
        on='contract_ticker',
        how='left',
    )
